"""Extract base level metrics"""
from functools import cache

import numpy as np
import opendssdirect as odd
import pandas as pd
import polars as pl

from emerge.network.asset_metrics import networkx_from_opendss_model


def get_allbus_voltage_pu():
    return odd.Circuit.AllBusMagPu()


def get_voltage_distribution(
    voltage_array: list, bins: list = [0.7 + i * 0.05 for i in range(12)]
):
    voltage_array = np.array(voltage_array)
    voltage_array = voltage_array[voltage_array != 0]

    hist, _ = np.histogram(voltage_array, bins)
    pct_hist = list(np.array(hist) * 100 / len(voltage_array))

    return {"Voltage bins": bins[:-1], "Percentage nodes": pct_hist}


def get_voltage_by_distance():
    mapper = {1: "Phase A", 2: "Phase B", 3: "Phase C"}
    voltage_by_distance = {"Voltage (pu)": [], "Distance from substation (km)": [], "Phase": []}
    for phase in range(1, 4):
        pu_voltages = odd.Circuit.AllNodeVmagPUByPhase(phase)
        node_distances = odd.Circuit.AllNodeDistancesByPhase(phase)

        voltage_by_distance["Voltage (pu)"].extend(pu_voltages)
        voltage_by_distance["Phase"].extend([mapper[phase]] * len(pu_voltages))
        voltage_by_distance["Distance from substation (km)"].extend(node_distances)

    return voltage_by_distance


def get_voltage_by_lat_lon():
    network = networkx_from_opendss_model()
    node_data = {node[0]: node[1] for node in network.nodes.data()}
    all_bus_voltage = odd.Circuit.AllBusMagPu()
    all_bus_names = odd.Circuit.AllBusNames()

    voltage_by_lat_lon = {"longitudes": [], "latitudes": [], "voltage (pu)": []}

    for bus, voltage_pu in zip(all_bus_names, all_bus_voltage):
        voltage_by_lat_lon["latitudes"].append(node_data[bus]["pos"][1])
        voltage_by_lat_lon["longitudes"].append(node_data[bus]["pos"][0])
        voltage_by_lat_lon["voltage (pu)"].append(voltage_pu)

    return voltage_by_lat_lon


@cache
def get_buses() -> list[str]:
    """Returns list of buses for all nodes in current openodd circuit."""
    nodes = odd.Circuit.AllNodeNames()
    return [el.split(".")[0] for el in nodes]


@cache
def get_branch_elements() -> list[str]:
    """Returns list of buses for all branches in current openodd circuit."""
    return odd.PDElements.AllNames()


def get_voltage_dataframe():
    """Function to retrieve voltage dataframe for all buses."""
    voltage_df = {"busname": get_buses(), "voltage(pu)": odd.Circuit.AllBusMagPu()}
    return pl.DataFrame(voltage_df)


def get_loading_dataframe():
    """Function to retrieve line loading dataframe for all line segments."""
    loading = odd.PDElements.AllPctNorm(AllNodes=False)
    return pl.DataFrame({"branch": get_branch_elements(), "loading(pu)": np.array(loading) / 100})


def get_pv_power_dataframe():
    """Function to retrieve pv power dataframe."""
    pv_power_df = {"pvname": [], "active_power": [], "reactive_power": []}

    flag = odd.PVsystems.First()
    while flag > 0:
        pv_name = odd.PVsystems.Name().lower()
        pv_powers = odd.CktElement.Powers()

        active_power = -sum(pv_powers[::2])
        reactive_power = -sum(pv_powers[1::2])
        pv_power_df["pvname"].append(pv_name)
        pv_power_df["active_power"].append(active_power)
        pv_power_df["reactive_power"].append(reactive_power)

        flag = odd.PVsystems.Next()

    return pd.DataFrame(pv_power_df).set_index("pvname")



""" OpenDSS simulator service for performing power flow."""

from pathlib import Path

import opendssdirect as dss
from loguru import logger


class OpenDSSSimulator:
    def __init__(
        self,
        path_to_master_dss_file,
    ):
        self.case_file = Path(path_to_master_dss_file)
        self.dss_instance = dss

        self.dss_instance.run_command("Clear")
        self.dss_instance.Basic.ClearAll()
        self.execute_dss_command(f"Redirect {self.case_file}")

    def execute_dss_command(self, dss_command: str):
        """Method to run opendss commands."""
        error = self.dss_instance.run_command(dss_command)
        if error:
            logger.error(f"Error executing command {dss_command} >> {error}")
            raise Exception(f"Error executing command {dss_command} >> {error}")
        logger.info(f"Sucessfully executed the command, {dss_command}")
        return error

    def set_frequency(self, frequency: int):
        """Method to set frequency."""
        self.execute_dss_command(f"Set DefaultBaseFrequency={frequency}")

    def set_mode(self, mode: int):
        """Method to set simulation mode.

        For example: 2 means QSTS mode.
        """
        self.dss_instance.Solution.Mode(mode)

    def set_max_iteration(self, max_iterations: int):
        """ "Method to set max iterations."""
        self.dss_instance.Solution.MaxControlIterations(max_iterations)
        self.dss_instance.Solution.MaxIterations(max_iterations)

    def set_stepsize(self, step_in_min: float):
        """Method to set step size."""
        self.dss_instance.Solution.StepSizeMin(step_in_min)

    def post_redirect(self, dss_file_path: Path):
        """Redirect this file."""
        self.execute_dss_command(f"Redirect {str(dss_file_path.absolute())}")
        self.recalc()
        self.solve()

    def recalc(self):
        """Method to recal and solve."""
        self.execute_dss_command("calcv")

    def solve(self):
        self.dss_instance.Solution.Number(1)
        self.dss_instance.Solution.Solve()
        return self.dss_instance.Solution.Converged()

    def set_simulation_time(self, sim_time, profile_start_time):
        time_diff = sim_time - profile_start_time
        total_seconds = time_diff.total_seconds()
        hours = int(total_seconds / 3600)
        seconds = total_seconds - hours * 3600
        self.dss_instance.Solution.Hour(hours)
        self.dss_instance.Solution.Seconds(seconds)



from typing import Optional
from pathlib import Path
from datetime import datetime

from sqlmodel import Field, SQLModel, create_engine


class HostingCapacityReport(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    node_name: str
    hosting_capacity_kw: float
    sardi_voltage: float
    sardi_line: float
    sardi_aggregated: float


class SimulationTime(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    node_name: str
    capacity: float
    compute_sec: float


class SimulationConvergenceReport(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    capacity_kw: float
    node_name: str
    convergence: bool
    timestamp: datetime


class TotalEnergyReport(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    node_name: str
    pv_capacity_kw: float
    pv_energy_mwh: float
    circuit_energy_mwh: float


class OverloadedLinesReport(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    start_time: datetime
    resolution_min: int
    node_name: str
    line_name: str
    loadings: str


def get_engine(sqlite_file: Path):
    return create_engine(f"sqlite:///{str(sqlite_file)}")


def create_table(sqlite_file: Path):
    """Function to create sqlite table."""

    engine = create_engine(f"sqlite:///{str(sqlite_file)}")
    SQLModel.metadata.create_all(engine)
    return engine



"""Extract base level metrics."""

import networkx as nx

import opendssdirect as odd
from emerge.network.data_model import (
    AssetMetrics,
    FeederMetrics,
    LoadAssetMetrics,
    PVAssetMetrics,
    CapacitorAssetMetrics,
    RegulatorsAssetMetrics,
    TransformersAssetMetrics,
)


def opendss_load_metrics_extractor():
    flag = odd.Loads.First()
    kws, kvars = [], []
    while flag:
        kws.append(odd.Loads.kW())
        kvars.append(odd.Loads.kvar())
        flag = odd.Loads.Next()

    return LoadAssetMetrics(
        total_count=odd.Loads.Count(),
        max_kw_capacity=round(max(kws), 3) if kws else 0,
        min_kw_capacity=round(min(kws), 3) if kws else 0,
        min_kvar_capacity=round(min(kvars), 3) if kvars else 0,
        max_kvar_capacity=round(max(kvars), 3) if kvars else 0,
        total_kw_capacity=round(sum(kws), 3) if kws else 0,
        total_kvar_capacity=round(sum(kvars), 3) if kvars else 0,
    )


def opendss_pv_metrics_extractor():
    flag = odd.PVsystems.First()
    kws = []
    while flag:
        kws.append(odd.PVsystems.Pmpp())
        flag = odd.PVsystems.Next()

    return PVAssetMetrics(
        total_count=odd.PVsystems.Count(),
        max_kw_capacity=round(max(kws), 3) if kws else 0,
        min_kw_capacity=round(min(kws), 3) if kws else 0,
        total_kw_capacity=round(sum(kws), 3) if kws else 0,
    )


def opendss_capacitors_metrics_extractor():
    flag = odd.Capacitors.First()
    kvars = []
    while flag:
        kvars.append(odd.Capacitors.kvar())
        flag = odd.PVsystems.Next()

    return CapacitorAssetMetrics(
        total_count=odd.Capacitors.Count(),
        max_kvar_capacity=round(max(kvars), 3) if kvars else 0,
        min_kvar_capacity=round(min(kvars), 3) if kvars else 0,
        total_kvar_capacity=round(sum(kvars), 3) if kvars else 0,
    )


def opendss_regulators_metrics_extractor():
    return RegulatorsAssetMetrics(total_count=odd.RegControls.Count())


def opendss_transformers_metrics_extractor():
    flag = odd.Transformers.First()
    kvas = []
    while flag:
        kvas.append(odd.Transformers.kVA())
        flag = odd.Transformers.Next()

    return TransformersAssetMetrics(
        total_count=odd.Transformers.Count(),
        max_kva_capacity=round(max(kvas), 3) if kvas else 0,
        min_kva_capacity=round(min(kvas), 3) if kvas else 0,
        total_kva_capacity=round(sum(kvas), 3) if kvas else 0,
    )


def networkx_from_opendss_model():
    """Let's create a networkx representation of the model"""
    network = nx.Graph()

    """ Let's add all buses """
    for bus in odd.Circuit.AllBusNames():
        odd.Circuit.SetActiveBus(bus)
        network.add_node(bus, pos=(odd.Bus.X(), odd.Bus.Y()))

    UNIT_MAPPER = {
        0: 0,
        1: 1.60934,
        2: 0.3048,
        3: 1,
        4: 0.001,
        5: 0.0003048,
        6: 0.0000254,
        7: 0.00001,
    }

    """ Let's all line elements to the network """

    for edge_name in ["Line", "Transformer"]:
        odd.Circuit.SetActiveClass(edge_name)
        flag = odd.ActiveClass.First()
        while flag > 0:
            name = odd.CktElement.Name().lower()
            buses = odd.CktElement.BusNames()
            edge_length = (
                0
                if edge_name == "Transformer"
                else UNIT_MAPPER[odd.Lines.Units()] * odd.Lines.Length()
            )

            network.add_edge(
                buses[0].split(".")[0], buses[1].split(".")[0], name=name, length=edge_length
            )
            flag = odd.ActiveClass.Next()

    return network


def get_all_kv_levels():
    flag = odd.Transformers.First()
    kvs = []
    while flag:
        for wdg in range(1, odd.Transformers.NumWindings() + 1):
            odd.Transformers.Wdg(wdg)
            kvs.append(odd.Transformers.kV())
        flag = odd.Transformers.Next()
    return list(set(kvs))


def opendss_feeder_metrics_extractor():
    df = odd.utils.class_to_dataframe("vsource").to_dict()
    source_bus = df["bus1"]["vsource.source"].split(".")[0]

    """ Get networkx representation of opendss model """
    network = networkx_from_opendss_model(odd)

    """ Let's create dfs tree """
    dfs_tree = nx.dfs_tree(network, source=source_bus)

    for edge in dfs_tree.edges():
        edge_data = network[edge[0]][edge[1]]
        dfs_tree[edge[0]][edge[1]]["length"] = edge_data["length"]

    total_feeder_length = nx.dag_longest_path_length(dfs_tree, weight="length", default_weight=0)
    longest_path_length = nx.dag_longest_path(dfs_tree, weight="length", default_weight=0)

    """ Get all the kvs """
    all_kvs = get_all_kv_levels(odd)
    all_kvs.sort()
    secondary_kv, primary_kv = all_kvs[0], all_kvs[1]

    transformer_buses = []

    flag = odd.Transformers.First()

    while flag:
        buses = odd.CktElement.BusNames()

        kvs = []
        for wdg in range(1, odd.Transformers.NumWindings() + 1):
            odd.Transformers.Wdg(wdg)
            kvs.append(odd.Transformers.kV())
        if secondary_kv in kvs:
            transformer_buses.append(buses[0].split(".")[0])
        flag = odd.Transformers.Next()

    secondary_lengths = {}

    for bus in transformer_buses:
        dfs_tree_ = nx.dfs_tree(dfs_tree, source=bus)
        for edge in dfs_tree_.edges():
            edge_data = network[edge[0]][edge[1]]
            dfs_tree_[edge[0]][edge[1]]["length"] = edge_data["length"]

        secondary_length = nx.dag_longest_path_length(dfs_tree_, weight="length", default_weight=0)
        secondary_lengths[bus] = secondary_length

    max_primary_feeder_length = total_feeder_length
    for bus, length in secondary_lengths.items():
        if bus in longest_path_length:
            max_primary_feeder_length = total_feeder_length - length

    secondary_lengths_list = list(secondary_lengths.values())

    return FeederMetrics(
        total_feeder_length_km=round(total_feeder_length, 3),
        max_primary_feeder_length_km=round(max_primary_feeder_length, 3),
        max_secondary_feeder_length_km=round(max(secondary_lengths_list), 3)
        if secondary_lengths_list
        else 0,
        min_secondary_feeder_length_km=round(min(secondary_lengths_list), 3)
        if secondary_lengths_list
        else 0,
        secondary_kv_level=secondary_kv,
        primary_kv_level=primary_kv,
        total_buses=odd.Circuit.NumBuses(),
        total_line_sections=odd.Lines.Count(),
    )


def aggregate_asset_metrics():
    loads = opendss_load_metrics_extractor()
    pvs = opendss_pv_metrics_extractor()
    capacitors = opendss_capacitors_metrics_extractor()
    regs = opendss_regulators_metrics_extractor()
    transformers = opendss_transformers_metrics_extractor()
    feeder = opendss_feeder_metrics_extractor()

    return AssetMetrics(
        loads=loads,
        pvs=pvs,
        capacitors=capacitors,
        regulators=regs,
        transformers=transformers,
        lines=feeder,
    )


def get_coordinates():
    network = networkx_from_opendss_model(odd)
    coordinates = {
        "nodes": {
            "latitudes": [],
            "longitudes": [],
        },
        "edges": {"latitudes": [], "longitudes": []},
    }

    node_data = {node[0]: node[1] for node in network.nodes.data()}
    for edge in network.edges():
        try:
            from_node = node_data[edge[0]]["pos"]
            to_node = node_data[edge[1]]["pos"]
            coordinates["nodes"]["latitudes"].extend([from_node[1], to_node[1]])
            coordinates["nodes"]["longitudes"].extend([from_node[0], to_node[0]])
            coordinates["edges"]["latitudes"].extend([from_node[1], to_node[1], None])
            coordinates["edges"]["longitudes"].extend([from_node[0], to_node[0], None])
        except Exception:
            print(f"{edge} can't get coord info {network.get_edge_data(*edge)}")

    return coordinates



""" Module for handling data model used in scenario generation. """


from typing import List, Optional, Dict
from enum import Enum

from pydantic import BaseModel, Field, model_validator
from pydantic_core import PydanticCustomError
from typing_extensions import Annotated, Literal


class CustomerModel(BaseModel):
    """Interface for representing customer used in pv scenario development."""

    name: str
    kw: float
    distance: float
    cust_type: Optional[str] = None


class DERType(str, Enum):
    """Available der enum types."""

    # pylint: disable=invalid-name
    load = "load"
    solar = "solar"


class CapacityStrategyEnum(str, Enum):
    """Available strategies for sizing solar systems."""

    # pylint: disable=invalid-name
    solar_energy_based = "solar_energy_based"
    peak_multiplier = "peak_multiplier"
    fixed_sizing = "fixed_sizing"


class SelectionStrategyEnum(str, Enum):
    """Available strategies for selecting loads to have pv system."""

    # pylint: disable=invalid-name
    random_allocation = "random_allocation"
    far_allocation = "far_allocation"
    near_allocation = "near_allocation"


class EnergyBasedSolarSizingStrategyInput(BaseModel):
    """Input model for default solar sizing strategy."""

    capacity_factor: Annotated[float, Field(gt=0, le=1)] = 0.3
    load_factor: Annotated[float, Field(gt=0, le=1)] = 0.33
    max_pct_production: Annotated[float, Field(gt=0, le=500)] = 100
    profile: str


class SizeWithProbabilityModel(BaseModel):
    """Use this model if you want to pick sizes with probabilities."""

    sizes: List[float] | float
    probabilites: List[Annotated[float, Field(ge=0, le=1)]] | float
    profile: List[str] | str

    @model_validator(mode="after")
    def validate_lengths_are_same(self) -> "SizeWithProbabilityModel":
        if (
            isinstance(self.sizes, list)
            or isinstance(self.probabilites, list)
            or isinstance(self.profile, list)
        ):
            try:
                if len(self.sizes) == len(self.probabilites) == len(self.profile):
                    return self
                raise ValueError("Length must match for sizes,  probabilities and profile names. ")
            except Exception as error:
                raise PydanticCustomError("mixed_type", f"Error validating fields : {error}")
        return self


class _DERScenarioInput(BaseModel):
    """Base input model for der scenarios."""

    sizing_strategy: CapacityStrategyEnum
    der_type: DERType
    der_tag: str = ""
    energy_sizing_input: Dict[
        str, EnergyBasedSolarSizingStrategyInput
    ] | EnergyBasedSolarSizingStrategyInput | None = None
    peakmult_sizing_input: Dict[
        str, SizeWithProbabilityModel
    ] | SizeWithProbabilityModel | None = None
    fixed_sizing_input: Dict[
        str, SizeWithProbabilityModel
    ] | SizeWithProbabilityModel | None = None


class DERScenarioInput(_DERScenarioInput):
    """Input model for der scenarios."""

    selection_strategy: SelectionStrategyEnum
    other_ders: List[_DERScenarioInput]


class BasicDERModel(BaseModel):
    """Basic DER model used in solar scenario development."""

    name: str
    kw: Annotated[float, Field(ge=0)]
    customer: CustomerModel
    der_type: DERType
    der_tag: str = ""
    profile: str


class DistDERScenarioModel(BaseModel):
    """Model for storing solars in a given scenario."""

    name: str
    sample_id: int
    penetration: float
    ders: List[BasicDERModel]


class LoadMetadataModel(BaseModel):
    """Interface for representing OpenDSS load metadata."""

    name: str
    bus: str
    num_phase: Annotated[int, Field(ge=1, le=3)]
    kv: Annotated[float, Field(gt=0)]
    yearly: Optional[str] = None


class DERScenarioInputModel(_DERScenarioInput):
    """Interface for der scenario input model."""

    file_name: str
    selection_strategy: SelectionStrategyEnum
    other_ders: List[_DERScenarioInput]


class ScenarioBaseConfig(BaseModel):
    """Interface for basic settings in defining scenario."""

    pct_resolution: Annotated[float, Field(gt=0, le=100)] = Field(
        description="Percentage resolution or step resolution"
    )
    num_of_penetration: Annotated[int, Field(gt=0)]
    max_num_of_samples: Annotated[int, Field(ge=1)] = 1


class DERScenarioConfigModel(ScenarioBaseConfig):
    """CLI interface model for generating solar scenarios."""

    master_file: str
    output_folder: str
    opendss_attr: Literal["yearly", "class"] = "class"
    der_scenario: List[DERScenarioInputModel]



"""
Create a geoJSON for a feeder model
"""

from pathlib import Path

from emerge.utils.util import validate_path, write_file


def create_feeder_geojson(dss_instance, output_folder):
    output_folder = Path(output_folder)
    output_folder.mkdir(exist_ok=True)
    validate_path(output_folder)

    # Let's get all the buses
    bus_geojson = {"type": "FeatureCollection", "features": []}

    # Get all the busnames, loop over them and find the
    # x,y coordinates
    all_buses = dss_instance.Circuit.AllBusNames()
    bus_coord_dict = {}
    for bus in all_buses:
        dss_instance.Circuit.SetActiveBus(bus)

        x, y = dss_instance.Bus.X(), dss_instance.Bus.Y()
        bus_coord_dict[bus] = {"longitude": x, "latitude": y}

        bus_geojson["features"].append(
            {
                "type": "Feature",
                "geometry": {"type": "Point", "coordinates": [x, y]},
                "properties": {"name": bus},
            }
        )
    write_file(bus_geojson, output_folder / "buses.json")

    # Get all the line sections
    lines_geojson = {"type": "FeatureCollection", "features": []}
    flag = dss_instance.Lines.First()
    while flag:
        line_name = dss_instance.Lines.Name()
        bus1, bus2 = (
            dss_instance.Lines.Bus1().split(".")[0],
            dss_instance.Lines.Bus2().split(".")[0],
        )
        n_customers = dss_instance.Lines.TotalCust()
        lines_geojson["features"].append(
            {
                "type": "Feature",
                "geometry": {
                    "type": "LineString",
                    "coordinates": [
                        [bus_coord_dict[bus1]["longitude"], bus_coord_dict[bus1]["latitude"]],
                        [bus_coord_dict[bus2]["longitude"], bus_coord_dict[bus2]["latitude"]],
                    ],
                    "properties": {
                        "name": line_name,
                        "n_customers": n_customers,
                        "bus1": bus1,
                        "bus2": bus2,
                    },
                },
            }
        )

        flag = dss_instance.Lines.Next()

    write_file(lines_geojson, output_folder / "lines.json")

    # Get all transformers
    transformer_geojson = {"type": "FeatureCollection", "features": []}
    flag = dss_instance.Transformers.First()
    while flag:
        transformer_name = dss_instance.CktElement.Name()
        bus1, bus2 = dss_instance.CktElement.BusNames()[:2]
        bus1, bus2 = bus1.split(".")[0], bus2.split(".")[0]
        transformer_geojson["features"].append(
            {
                "type": "Feature",
                "geometry": {
                    "type": "Point",
                    "coordinates": [
                        bus_coord_dict[bus1]["longitude"],
                        bus_coord_dict[bus1]["latitude"],
                    ],
                    "properties": {
                        "name": transformer_name,
                        "bus1": bus1,
                        "bus2": bus2,
                        "kva": dss_instance.Transformers.kVA(),
                    },
                },
            }
        )

        flag = dss_instance.Transformers.Next()

    write_file(transformer_geojson, output_folder / "transformers.json")

    # Get all loads
    load_geojson = {"type": "FeatureCollection", "features": []}
    flag = dss_instance.Loads.First()
    while flag:
        load_name = dss_instance.CktElement.Name()
        bus1 = dss_instance.CktElement.BusNames()[0]
        bus1 = bus1.split(".")[0]
        load_geojson["features"].append(
            {
                "type": "Feature",
                "geometry": {
                    "type": "Point",
                    "coordinates": [
                        bus_coord_dict[bus1]["longitude"],
                        bus_coord_dict[bus1]["latitude"],
                    ],
                    "properties": {
                        "name": load_name,
                        "bus": bus1,
                        "kw": dss_instance.Loads.kW(),
                        "kvar": dss_instance.Loads.kvar(),
                    },
                },
            }
        )

        flag = dss_instance.Loads.Next()

    write_file(load_geojson, output_folder / "loads.json")



""" Module for managing EMERGE specific exceptions. """


class BaseEmergeException(Exception):
    """Base class for managing emerge exceptions."""


class EnergyMeterNotDefined(Exception):
    """Exception raised because energy meter is not defined."""



"""This module implements cli function to
run powerflow on single opendss model."""

from datetime import datetime
from pathlib import Path
from typing import Annotated
import json
from emerge.simulator.opendss import OpenDSSSimulator

from pydantic import BaseModel, Field
import click

from emerge.cli import get_observers
from emerge.cli.get_observers import SimulationMetrics
from emerge.metrics import observer
from emerge.simulator.simulation_manager import OpenDSSSimulationManager


class TimeseriesSimulationInput(BaseModel):
    """Interface for timeseries simulation input model."""

    master_dss_file: Annotated[Path, Field(..., description="Path to master dss file.")]
    start_time: Annotated[datetime, Field(..., description="Start time for simulation.")]
    end_time: Annotated[datetime, Field(..., description="End time for simulation.")]
    profile_start_time: Annotated[datetime, Field(..., description="Profile start time.")]
    export_path: Annotated[Path, Field(..., description="Path to export metric results.")]
    resolution_min: Annotated[
        float, Field(60, gt=0, description="Simulation time resolution in minute.")
    ]
    metrics: Annotated[
        SimulationMetrics, Field(SimulationMetrics(), description="Simulation metrics.")
    ]


def compute_timeseries_simulation_metrics(config: TimeseriesSimulationInput):
    """Function to compute metrics for timeseries simulation."""

    opendss_instance = OpenDSSSimulator(config.master_dss_file)

    manager = OpenDSSSimulationManager(
        opendss_instance=opendss_instance,
        simulation_start_time=config.start_time,
        profile_start_time=config.profile_start_time,
        simulation_end_time=config.end_time,
        simulation_timestep_min=config.resolution_min,
    )
    subject = observer.MetricsSubject()
    observers = get_observers.get_observers(config.metrics.model_dump())

    for _, observer_ in observers.items():
        subject.attach(observer_)

    manager.simulate(subject)
    manager.export_convergence(config.export_path / "convergence_report.csv")
    observer.export_csv(list(observers.values()), config.export_path)


@click.command()
@click.option(
    "-c",
    "--config",
    help="Path to config file for generating scenarios",
)
def timeseries_simulation(config):
    """Function to run timeseries simulation."""

    with open(config, "r", encoding="utf-8") as file:
        config_dict = json.load(file)

    config = TimeseriesSimulationInput.model_validate(config_dict)
    compute_timeseries_simulation_metrics(config)



""" Module for managing metric computation subscriber and publisher. """

import abc
import uuid
from typing import Dict, List
import polars


class MetricObserver(abc.ABC):
    """Abstracte interace for metrics observers"""

    _id = str(uuid.uuid4())

    @abc.abstractmethod
    def compute(self, *args, **kwargs) -> None:
        """All metric observer subclass must implement compute method."""

    @abc.abstractmethod
    def get_metric(self) -> Dict:
        """All metric observer subclass must implement get_metric method."""


class MetricsSubject:
    """Class for managing metric subscribers"""

    def __init__(self):
        self.subscribers: list[MetricObserver] = []

    def _observer_exists(self, observer: MetricObserver):
        """Check whether the observer exists or not"""

        for id, obs in enumerate(self.subscribers):
            if obs._id == observer._id:
                return id

        return False

    def attach(self, observer: MetricObserver):
        """Method for attaching the observers."""
        if not self._observer_exists(observer):
            self.subscribers.append(observer)

    def detach(self, observer: MetricObserver):
        """Method for deleting the observer object from the list."""

        observer_index = self._observer_exists(observer)
        if observer_index:
            self.subscribers.pop(observer_index)

    def notify(self):
        """Method for notifying the observers."""
        for obs in self.subscribers:
            obs.compute()


def export_csv(observers: List[MetricObserver], export_path):
    for observer in observers:
        df = polars.from_dict(observer.get_metric())
        df.write_csv(export_path / (str(observer.__class__.__name__) + ".csv"))



"""This module contains function to get observers."""

from typing import Annotated, Any, Literal

from pydantic import Field, BaseModel

from emerge.metrics import system_metrics
from emerge.metrics import node_voltage_stats
from emerge.metrics import line_loading_stats


CLASS_MAPPER = {
    "substation_power": system_metrics.TimeseriesTotalPower,
    "total_pvpower": system_metrics.TimeseriesTotalPVPower,
    "total_powerloss": system_metrics.TimeseriesTotalLoss,
    "node_timeseries_voltage": node_voltage_stats.NodeVoltageTimeSeries,
    "node_voltage_stats": node_voltage_stats.NodeVoltageStats,
    "node_voltage_bins": node_voltage_stats.NodeVoltageBins,
    "line_loading_stats": line_loading_stats.LineLoadingStats,
    "line_loading_bins": line_loading_stats.LineLoadingBins,
    "overloaded_lines": line_loading_stats.OverloadedLines,
    "sardi_voltage": system_metrics.SARDI_voltage,
    "sardi_line": system_metrics.SARDI_line,
    "sardi_aggregated": system_metrics.SARDI_aggregated,
}


class MetricEntry(BaseModel):
    """Interface for defining metric entry."""

    export_type: Annotated[
        Literal["csv"],
        Field("csv", description="Export format type. Only csv is supported for now."),
    ]
    file_name: Annotated[str, Field(..., description="File name with extension.")]
    args: Annotated[list[Any], Field([], description="List of arguments if any.")]


class SimulationMetrics(BaseModel):
    """Interface for defining simulation metrics."""

    node_timeseries_voltage: Annotated[
        MetricEntry,
        Field(
            MetricEntry(file_name="node_voltage_timeseries.csv"),
            description="Statistics for node time series voltages.",
        ),
    ]

    node_voltage_stats: Annotated[
        MetricEntry,
        Field(
            MetricEntry(file_name="node_voltage_stats.csv"),
            description="Statistics for node voltages.",
        ),
    ]
    node_voltage_bins: Annotated[
        MetricEntry,
        Field(
            MetricEntry(
                file_name="node_voltage_bins.csv",
                args=[[0.95, 0.96, 0.97, 0.98, 0.99, 1.0, 1.01, 1.02, 1.03, 1.04, 1.05]],
            ),
            description="Distribution for node voltages.",
        ),
    ]
    line_loading_stats: Annotated[
        MetricEntry,
        Field(
            MetricEntry(file_name="line_loading_stats.csv"),
            description="Statistics about thermal loading of line segments.",
        ),
    ]
    line_loading_bins: Annotated[
        MetricEntry,
        Field(
            MetricEntry(
                file_name="line_loading_bins.csv",
                args=[[0.2, 0.4, 0.6, 0.8, 1.0, 1.2, 1.4, 1.6, 2]],
            ),
            description="Distribution of thermal loading of line segments.",
        ),
    ]
    xfmr_loading_stats: Annotated[
        MetricEntry,
        Field(
            MetricEntry(file_name="xfmr_loading_stats.csv"),
            description="Statistics about thermal loading of distribution transformers.",
        ),
    ]
    xfmr_loading_bins: Annotated[
        MetricEntry,
        Field(
            MetricEntry(
                file_name="xfmr_loading_bins.csv",
                args=[[0.2, 0.4, 0.6, 0.8, 1.0, 1.2, 1.4, 1.6, 2]],
            ),
            description="Distribution of thermal loading of distribution transformers.",
        ),
    ]
    substation_power: Annotated[
        MetricEntry,
        Field(
            MetricEntry(file_name="substation_power_timeseries.csv"),
            description="Time series substation power.",
        ),
    ]

    total_pvpower: Annotated[
        MetricEntry,
        Field(
            MetricEntry(file_name="total_pvpower_timeseries.csv"),
            description="Time series total solar power.",
        ),
    ]
    total_powerloss: Annotated[
        MetricEntry,
        Field(
            MetricEntry(file_name="total_loss_timeseries.csv"),
            description="Time series total power loss.",
        ),
    ]
    sardi_voltage: Annotated[
        MetricEntry,
        Field(
            MetricEntry(file_name="sardi_voltage.csv", args=[1.05, 0.95]),
            description="System average risk duration index associated with voltage violation.",
        ),
    ]
    sardi_line: Annotated[
        MetricEntry,
        Field(
            MetricEntry(file_name="sardi_line.csv", args=[1]),
            description="System average risk duration index associated with line thermal violation.",
        ),
    ]
    sardi_transformer: Annotated[
        MetricEntry,
        Field(
            MetricEntry(file_name="sardi_transformer.csv", args=[1]),
            description="System average risk duration index associated with transformer thermal violation.",
        ),
    ]
    sardi_aggregated: Annotated[
        MetricEntry,
        Field(
            MetricEntry(
                file_name="sardi_aggregated.csv",
                args=[1, {"overvoltage_threshold": 1.05, "undervoltage_threshold": 0.95}],
            ),
            description="System average risk duration index for all violations.",
        ),
    ]
    overloaded_lines: Annotated[
        MetricEntry,
        Field(
            MetricEntry(
                file_name="overloaded_lines.csv",
            ),
            description="Overloaded line along with their loadings.",
        ),
    ]
    overloaded_transformers: Annotated[
        MetricEntry,
        Field(
            MetricEntry(
                file_name="overloaded_transformers.csv",
            ),
            description="Overloaded transformers along with their loadings.",
        ),
    ]


def get_observers(metrics: dict) -> dict:
    """Function to return list of observers."""
    observers = {}

    for key, subdict in metrics.items():
        if key in CLASS_MAPPER:
            if "args" not in subdict:
                observers[key] = CLASS_MAPPER[key]()
            else:
                observers[key] = CLASS_MAPPER[key](*subdict["args"])
    return observers



""" Utility functions for odd. """

import copy
from typing import List

import opendssdirect as odd
import pandas as pd
import networkx as nx

from emerge.metrics.exceptions import EnergyMeterNotDefined
from emerge.network import asset_metrics
from emerge.scenarios import data_model


def get_bus_distance_dataframe() -> pd.DataFrame:
    """Get bus distance dataframe.

    Returns
    -------
        pd.DataFrame: Dataframe containing distance from
            substation indexed by busname.

    """

    bus_distance_df = {"busname": [], "distance": []}
    for bus in odd.Circuit.AllBusNames():
        odd.Circuit.SetActiveBus(bus)

        bus_distance_df["busname"].append(bus)
        bus_distance_df["distance"].append(odd.Bus.Distance())

    if not sum(bus_distance_df["distance"]):
        raise EnergyMeterNotDefined(
            "Define energy meter to be able to find"
            "number of customers downward of line segment."
        )
    return pd.DataFrame(bus_distance_df).set_index("busname")


def get_list_of_customer_models(
    load_multiplier=1.0, cust_type: str = "yearly"
) -> List[data_model.CustomerModel]:
    """Returns list of customer models from odd instance.

    Args:
        odd (odd): Instance of OpenoddDirect

    Returns:
        List[data_model.CustomerModel]: List of customer data
            model.
    """
    cust_type_func = {"class": odd.Loads.Class, "yearly": odd.Loads.Yearly}.get(cust_type)

    flag = odd.Loads.First()
    bus_distance_df = get_bus_distance_dataframe()
    customer_model = []
    while flag:
        busname = odd.CktElement.BusNames()[0].split(".")[0]
        customer_model.append(
            data_model.CustomerModel(
                name=odd.CktElement.Name(),
                kw=odd.Loads.kW() * load_multiplier,
                distance=bus_distance_df.loc[busname]["distance"],
                cust_type=str(cust_type_func()),
            )
        )

        flag = odd.Loads.Next()

    return customer_model


def get_load_mapper_objects() -> List[data_model.LoadMetadataModel]:
    """Returns list of load mapper object models.

    Args:
        odd (odd): Instance of OpenoddDirect

    Returns:
        List[data_model.LoadMetadataModel]: List of load
            metadata model.
    """

    flag = odd.Loads.First()
    mapper_model = []
    while flag:
        mapper_model.append(
            data_model.LoadMetadataModel(
                name=odd.CktElement.Name(),
                bus=odd.CktElement.BusNames()[0],
                num_phase=odd.Loads.Phases(),
                kv=odd.Loads.kV(),
                yearly=odd.Loads.Yearly(),
            )
        )

        flag = odd.Loads.Next()

    return mapper_model


def get_bus_load_dataframe() -> pd.DataFrame:
    """Bus to load mapping dataframe.

    Returns
    -------
        pd.DataFrame: Dataframe containing mapping between load
            name and bus name
    """

    flag = odd.Loads.First()
    load_bus_name_dict = {"busname": [], "loadname": []}
    while flag:
        load_name = odd.CktElement.Name()
        busname = odd.CktElement.BusNames()[0].split(".")[0]
        load_bus_name_dict["busname"].append(busname)
        load_bus_name_dict["loadname"].append(load_name)

        flag = odd.Loads.Next()

    return pd.DataFrame(load_bus_name_dict)


def get_bus_load_flag() -> pd.DataFrame:
    """Bus to load mapping dataframe.

    Returns
    -------
        pd.DataFrame: Dataframe containing load flag
            indexed by bus name.
    """

    load_bus_df = get_bus_load_dataframe()
    buses_with_load = list(load_bus_df["busname"])
    all_buses = odd.Circuit.AllBusNames()
    buses_without_load = list(set(all_buses) - set(buses_with_load))
    is_bus_load = {
        "busname": buses_with_load + buses_without_load,
        "is_load": [1] * len(buses_with_load) + [0] * len(buses_without_load),
    }

    return pd.DataFrame(is_bus_load).set_index("busname")


def get_line_customers() -> pd.DataFrame:
    """Function to retrieve dataframe containing number
    of downward serving customers for all line segments.

    Returns
    -------
        pd.DataFrame: Dataframe containing number of customers
            indexed by line name.
    """
    line_customers_df = {"linename": [], "customers": []}

    odd.Circuit.SetActiveClass("Line")
    flag = odd.ActiveClass.First()

    while flag:
        line_name = odd.CktElement.Name().lower()
        n_customers = odd.Lines.TotalCust()
        line_customers_df["linename"].append(line_name)
        line_customers_df["customers"].append(n_customers)
        flag = odd.ActiveClass.Next()

    if not any(line_customers_df["customers"]):
        raise EnergyMeterNotDefined(
            "Define energy meter to be able to find"
            "number of customers downward of line segment."
        )
    return pd.DataFrame(line_customers_df).set_index("linename")


def get_transformer_customers() -> pd.DataFrame:
    """Function to retrieve dataframe containing number
    of downward serving customers for all transformers.

    Returns
    -------
        pd.DataFrame: Dataframe containing number of customers
            indexed by transformer name.
    """
    xfmr_customers_df = {"transformername": [], "customers": []}

    network = asset_metrics.networkx_from_opendss_model()
    substation_bus = get_source_node()
    bus_load_flag_df = get_bus_load_flag()

    odd.Circuit.SetActiveClass("Transformer")
    flag = odd.ActiveClass.First()

    while flag:
        network_copy = copy.deepcopy(network)
        xmfr_name = odd.CktElement.Name().lower()

        edge_to_be_removed = []
        for edge in network_copy.edges():
            edge_data = network_copy.get_edge_data(*edge)
            if edge_data and "name" in edge_data and edge_data["name"] == xmfr_name:
                edge_to_be_removed.append(edge)

        network_copy.remove_edges_from(edge_to_be_removed)
        connected_buses = nx.node_connected_component(network_copy, substation_bus)

        filtered_bus_df = bus_load_flag_df.loc[bus_load_flag_df.index.difference(connected_buses)]

        n_customers = filtered_bus_df.sum()["is_load"]
        xfmr_customers_df["transformername"].append(xmfr_name)
        xfmr_customers_df["customers"].append(n_customers)
        flag = odd.ActiveClass.Next()

    return pd.DataFrame(xfmr_customers_df).set_index("transformername")


def get_source_node() -> str:
    """Function to return source node for odd model.

    Returns
    -------
        str: Name of the source node
    """

    df = odd.utils.class_to_dataframe("vsource")
    source_node = df["bus1"].tolist()[0].split(".")[0]
    return source_node



""" This module contains function that would compute nodal hosting
capacity.

Idea is to loop through all the nodes. Increase the solar capacity by step kw specified
by user and return max capacity for which the risk would be zero.
"""
from pathlib import Path
from typing import Annotated
from datetime import datetime
import json
import multiprocessing
import time
import math

import click
import numpy as np
import polars as pl
from pydantic import BaseModel, Field
from loguru import logger
from sqlmodel import Session

from emerge.cli import get_num_core
from emerge.metrics import observer
from emerge.simulator import opendss
from emerge.cli.nodal_hosting_sqlite_tables import (
    HostingCapacityReport,
    OverloadedLinesReport,
    SimulationConvergenceReport,
    SimulationTime,
    TotalEnergyReport,
    create_table,
    get_engine,
)
from emerge.metrics.line_loading_stats import OverloadedLines
from emerge.metrics.system_metrics import (
    SARDI_aggregated,
    SARDI_line,
    SARDI_voltage,
    TotalEnergy,
    TotalPVGeneration,
)
from emerge.simulator.simulation_manager import OpenDSSSimulationManager


class BasicSimulationSettings(BaseModel):
    """Interface for basic simulation settings."""

    master_dss_file: Annotated[Path, Field(..., description="Path to master dss file.")]
    start_time: Annotated[datetime, Field(..., description="Start time for simulation.")]
    end_time: Annotated[datetime, Field(..., description="End time for simulation.")]
    profile_start_time: Annotated[datetime, Field(..., description="Profile start time.")]
    resolution_min: Annotated[
        float, Field(60, gt=0, description="Simulation time resolution in minute.")
    ]


class SingleNodeHostingCapacityInput(BasicSimulationSettings):
    """Interface for single node hosting capacity."""

    step_kw: Annotated[
        float, Field(gt=0, description="kW value to increase pv capacity by each time.")
    ]
    max_kw: Annotated[float, Field(gt=0, description="Maximum kw to not exceed.")]


class MultiNodeHostingCapacityInput(SingleNodeHostingCapacityInput):
    """Interface for timeseries simulation input model."""

    num_core: Annotated[
        int, Field(ge=1, description="Number of cores to use for parallel simulation.")
    ]
    export_sqlite_path: Annotated[
        Path, Field(..., description="Sqlite file to export nodal hosting capacity.")
    ]
    pv_profile: Annotated[str, Field(..., description="Name of the profile for pv system")]


class NodalHostingCapacityReport:
    """Class interface for capturing nodal hosting capacity results."""

    def __init__(self):
        self.subject = observer.MetricsSubject()
        self.sardi_agg = SARDI_aggregated()
        self.sardi_voltage = SARDI_voltage()
        self.sardi_line = SARDI_line()
        self.solar_generation = TotalPVGeneration()
        self.ol_lines = OverloadedLines()
        self.circuit_energy = TotalEnergy()
        self.export_energy = TotalEnergy(export_only=True)

        for obs in [
            self.sardi_agg,
            self.sardi_voltage,
            self.sardi_line,
            self.solar_generation,
            self.ol_lines,
            self.circuit_energy,
            self.export_energy,
        ]:
            self.subject.attach(obs)

    def is_hosting_capacity_reached(self) -> bool:
        """Method to check if hosting capacity is reached."""
        return self.get_sardi_aggregated() > 0 or self.get_total_export_energy() != 0

    def get_subject(self) -> observer.MetricsSubject:
        """Return subject container containing observers."""
        return self.subject

    def get_overloaded_line_loadings(self) -> dict[str, list[float]]:
        """Returns overloaded lines."""
        return self.ol_lines.get_metric()

    def get_sardi_voltage(self):
        """Returns SARDI voltage metrics."""
        return pl.from_dict(self.sardi_voltage.get_metric())["sardi_voltage"].to_list()[0]

    def get_sardi_aggregated(self):
        """Method to return SARDI aggregate metric."""
        return pl.from_dict(self.sardi_agg.get_metric())["sardi_aggregated"].to_list()[0]

    def get_sardi_line(self):
        """Method to return SARDI line metric."""
        return pl.from_dict(self.sardi_line.get_metric())["sardi_line"].to_list()[0]

    def get_solar_total_energy(self):
        """Method to return total solar energy."""
        return pl.from_dict(self.solar_generation.get_metric())["active_power"].to_list()[0]

    def get_circuit_total_energy(self):
        """Method to get circuit total energy."""
        return pl.from_dict(self.circuit_energy.get_metric())["active_power"].to_list()[0]

    def get_total_export_energy(self):
        """Method to get total export energy."""
        return pl.from_dict(self.export_energy.get_metric())["active_power"].to_list()[0]


def _compute_hosting_capacity(input):
    """Wrapper around compute hosting capacity."""
    return compute_hosting_capacity(*input)


def compute_hosting_capacity(
    config: SingleNodeHostingCapacityInput, bus: str, pv_profile: str, sqlite_file: Path
):
    """Function to compute node hosting capacity."""
    hosting_capacity = 0
    engine = get_engine(sqlite_file)

    for capacity in np.arange(config.step_kw, config.max_kw, config.step_kw):
        opendss_instance = opendss.OpenDSSSimulator(config.master_dss_file)
        opendss_instance.dss_instance.Circuit.SetActiveBus(bus)
        bus_kv = round(opendss_instance.dss_instance.Bus.kVBase() * math.sqrt(3), 2)
        pv_name = f"{bus}_pv"

        new_pv = (
            f"new PVSystem.{pv_name} bus1={bus} kv={round(bus_kv,2 )} "
            + f"phases=3 kVA={capacity} Pmpp={capacity} PF=1.0 yearly={pv_profile}"
        )
        logger.info(new_pv)

        opendss_instance.dss_instance.run_command(new_pv)
        opendss_instance.recalc()
        opendss_instance.solve()

        sim_manager = OpenDSSSimulationManager(
            opendss_instance=opendss_instance,
            simulation_start_time=config.start_time,
            profile_start_time=config.profile_start_time,
            simulation_end_time=config.end_time,
            simulation_timestep_min=config.resolution_min,
        )

        report_instance = NodalHostingCapacityReport()
        logger.info(f"Node started {bus}, capacity {capacity}")

        start_time = time.time()
        sim_manager.simulate(subject=report_instance.get_subject())
        end_time = time.time()
        logger.info(f"Node finished {bus}, " f"elpased time {end_time - start_time} seconds")

        with Session(engine) as session:
            for timestamp, conv_ in zip(
                sim_manager.convergence_dict["datetime"],
                sim_manager.convergence_dict["convergence"],
            ):
                session.add(
                    SimulationConvergenceReport(
                        node_name=bus, convergence=conv_, capacity_kw=capacity, timestamp=timestamp
                    )
                )

            session.add(
                SimulationTime(
                    node_name=bus, capacity=capacity, compute_sec=(end_time - start_time)
                )
            )

            session.add(
                TotalEnergyReport(
                    node_name=bus,
                    pv_capacity_kw=capacity,
                    pv_energy_mwh=report_instance.get_solar_total_energy(),
                    circuit_energy_mwh=report_instance.get_circuit_total_energy(),
                )
            )

            for ol_line, loadings in report_instance.get_overloaded_line_loadings().items():
                session.add(
                    OverloadedLinesReport(
                        start_time=config.start_time,
                        resolution_min=config.resolution_min,
                        node_name=bus,
                        line_name=ol_line,
                        loadings=str(loadings),
                    )
                )

            session.commit()

        if report_instance.is_hosting_capacity_reached() > 0:
            break

        hosting_capacity = capacity

    with Session(engine) as session:
        session.add(
            HostingCapacityReport(
                node_name=bus,
                hosting_capacity_kw=hosting_capacity,
                sardi_voltage=report_instance.get_sardi_voltage(),
                sardi_aggregated=report_instance.get_sardi_aggregated(),
                sardi_line=report_instance.get_sardi_line(),
            )
        )
        session.commit()


@click.command()
@click.option(
    "-c",
    "--config",
    help="Path to config file for running timeseries simulation",
)
@click.option(
    "-n",
    "--nodes",
    default="",
    show_default=True,
    help="List of comma separated nodes",
)
def nodal_hosting_analysis(config: str, nodes: str):
    """Run multiscenario time series simulation and compute
    time series metrics."""

    analysis_start_time = time.time()
    with open(config, "r", encoding="utf-8") as file:
        config_dict = json.load(file)

    config: MultiNodeHostingCapacityInput = MultiNodeHostingCapacityInput.model_validate(
        config_dict
    )

    if not nodes:
        opendss_instance = opendss.OpenDSSSimulator(config.master_dss_file)
        buses = opendss_instance.dss_instance.Circuit.AllBusNames()
    else:
        buses = nodes.split(",")

    create_table(config.export_sqlite_path)
    num_core = get_num_core.get_num_core(config.num_core, len(buses))
    with multiprocessing.Pool(int(num_core)) as pool:
        data_to_process = [
            [
                SingleNodeHostingCapacityInput.model_validate(config.model_dump()),
                bus,
                config.pv_profile,
                config.export_sqlite_path,
            ]
            for bus in buses
        ]
        pool.map(_compute_hosting_capacity, data_to_process)
    analysis_end_time = time.time()
    print(f"Required time: {analysis_end_time - analysis_start_time} seconds, num_core={num_core}")



""" Basic utility functions. """

from typing import List, Dict, Union
from pathlib import Path
import json
import json5
import time

import yaml

from loguru import logger


def validate_path(
    path: str,
    check_for_dir: bool = True,
    check_for_file: bool = False,
    file_types: Union[None, List[str]] = None,
) -> None:
    """Check for path validation.

    Args:
        path (str): Path for either file or folder
        check_for_dir (bool): Set to true if need to check for directory
        check_for_file (bool): Set to true if need to check of
            existence of file
        file_types (Union[None, List[str]]): List of file types against
            which file path is to be checked.

    Raises:
        Exception: If file path does not exist or folder path does
            not exist or passed file type do not match what is in
            file_types list.
    """

    if not file_types:
        file_types = []

    path = Path(path)

    if not path.exists():
        raise Exception(f"{path} does not exist!")

    if check_for_dir and not path.is_dir():
        raise Exception(f"{path} is not a directory!")

    if check_for_file and path.suffix not in file_types:
        raise Exception(
            f"Invalid file type passed {path} , \
            valid file types are {file_types}"
        )


def read_file(file_path: str, use_json5=False) -> Dict:
    """Reads a file and returns a python dictionary."""

    file_path = Path(file_path)
    logger.debug(f"Attempting to read {file_path}")

    validate_path(
        file_path, check_for_dir=False, check_for_file=True, file_types=[".json", ".yaml"]
    )

    if file_path.suffix == ".json":
        with open(file_path, "r") as f:
            content = json.load(f) if not use_json5 else json5.load(f)

    elif file_path.suffix == ".yaml":
        with open(file_path, "r") as f:
            content = yaml.safe_load(f)

    else:
        logger.error(
            f"""Could not read the {file_path}, this feature is not yet
        implemented"""
        )
        raise Exception(
            f"File of type {file_path.suffix} \
            is not yet implemented for reading purpose"
        )

    logger.debug(f"{file_path} read successfully")
    return content


def write_file(content: dict, file_path: str, use_json5=False, **kwargs) -> None:
    """Takes a python dictionary and writes it into a file."""
    file_path = Path(file_path)
    validate_path(file_path.parent)

    if file_path.suffix == ".json":
        with open(file_path, "w") as f:
            json.dump(content, f, **kwargs) if not use_json5 else json5.dump(content, f, **kwargs)

    elif file_path.suffix == ".yaml":
        with open(file_path, "w") as f:
            yaml.safe_dump(content, f, **kwargs)

    else:
        raise Exception(
            f"File of type {file_path.suffix} \
            is not yet implemented for writing purpose"
        )


def time_tracker(func):
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        elapsed_time = end_time - start_time
        logger.info(f" '{func.__name__}' took {elapsed_time:.6f}s to execute.")
        return result

    return wrapper



""" Module for managing pydantic data models."""

from typing_extensions import Annotated

from pydantic import Field, BaseModel


class ThermalLoadingLimit(BaseModel):
    """Model representing thermal loading limit.
    Example:

        >>> from emerge.metrics.data_model import ThermalLoadingLimit
        >>> ThermalLoadingLimit(threshold=1.0)
    """

    threshold: Annotated[float, Field(ge=0.0, le=2.0)] = 1.0


class VoltageViolationLimit(BaseModel):
    """Model representing voltage violation limit."""

    overvoltage_threshold: Annotated[float, Field(ge=1.01, le=2.0)] = 1.05
    undervoltage_threshold: Annotated[float, Field(ge=0, le=1.0)] = 0.95



""" This module manages time series simulation for computing
time series metrics. """

from typing import Union
import datetime

import pandas as pd
from loguru import logger

from emerge.constants import OPENDSS_MAX_ITERATION, OPENDSS_QSTS_MODE
from emerge.simulator import opendss
from emerge.metrics import observer


class OpenDSSSimulationManager:
    """Class for managing time series simulation using OpenDSS.

    Parameters
    ----------

    opendss_instance :opendss.OpenDSSSimulator
        OpenDSS simulator instance.
    simulation_start_time: datetime.datetime
        Datetime indicating simulation start time.
    profile_start_time :datetime.datetime
        Datetime indicating profile start time.
    simulation_end_time :datetime.datetime
        Datetime indicating when the simulation finishes.
    simulation_timestep_min : float
        Simulation time resolution in minutes.
    """

    def __init__(
        self,
        opendss_instance: opendss.OpenDSSSimulator,
        simulation_start_time: datetime.datetime,
        profile_start_time: datetime.datetime,
        simulation_end_time: datetime.datetime,
        simulation_timestep_min: float,
    ) -> None:
        self.simulation_start_time = simulation_start_time
        self.profile_start_time = profile_start_time
        self.simulation_end_time = simulation_end_time
        self.simulation_timestep_min = simulation_timestep_min
        self.opendss_instance = opendss_instance

        self.opendss_instance.set_mode(OPENDSS_QSTS_MODE)
        self.opendss_instance.set_simulation_time(
            self.simulation_start_time, self.profile_start_time
        )
        self.opendss_instance.set_stepsize(self.simulation_timestep_min)
        self.opendss_instance.set_max_iteration(OPENDSS_MAX_ITERATION)
        self.current_time = self.simulation_start_time
        self.convergence_dict = {"datetime": [], "convergence": []}

    def update_convergence_dict(self, current_time: datetime.datetime, convergence: bool):
        """Updates convergence dict."""

        if not convergence:
            self.convergence_dict["datetime"].append(current_time)
            self.convergence_dict["convergence"].append(convergence)

    def export_convergence(self, export_path: str = "./convergence.csv"):
        """Export the convergence."""

        export_df = pd.DataFrame(self.convergence_dict)
        export_df.to_csv(export_path)

    def simulate(self, subject: Union[observer.MetricsSubject, None] = None):
        """Loops through all the simulation timesteps"""
        while self.current_time <= self.simulation_end_time:
            convergence = self.opendss_instance.solve()
            self.update_convergence_dict(self.current_time, convergence)

            if subject:
                subject.notify()
            self.current_time += datetime.timedelta(minutes=self.simulation_timestep_min)
            if not convergence:
                logger.error(f"Simulation finished for {self.current_time} >> {convergence}")



""" Module for managing computation of system level metrics. """
import networkx as nx
import opendssdirect as odd
import polars as pl

from emerge.metrics import observer
from emerge.simulator import powerflow_results
from emerge.utils import dss_util
from emerge.metrics import data_model
from emerge.network import asset_metrics


def _get_unimpacted_buses(graph: nx.Graph, impacted_edges: list[str], source_bus: str):
    """Internal function to return unimpacted buses."""

    graph_copy = nx.Graph()
    edge_to_be_removed = []
    for edge in graph.edges():
        graph_copy.add_edge(*edge)
        edge_data = graph.get_edge_data(*edge)
        if edge_data and "name" in edge_data and edge_data["name"] in impacted_edges:
            edge_to_be_removed.append(edge)

    graph_copy.remove_edges_from(edge_to_be_removed)
    connected_buses = nx.node_connected_component(graph_copy, source_bus)
    return connected_buses


def _get_voltage_impacted_buses(voltage_df: pl.DataFrame, ov_th: float, uv_th: float) -> list[str]:
    """Internal function to return list of buses impacted by voltage violations."""
    v_col = pl.col("voltage(pu)")
    return (
        voltage_df.filter((v_col > ov_th) & (v_col < uv_th) & (v_col > 0))
        .group_by("busname")
        .agg(v_col.mean())["busname"]
        .to_list()
    )


class TimeseriesTotalLoss(observer.MetricObserver):
    """Class for computing total loss.

    Attributes:
        active_power (float): Time series active power
        reactive_power (float): Time series reactive power
    """

    def __init__(self):
        self.active_power = []
        self.reactive_power = []

    def compute(self):
        """Refer to base class for more details."""
        timestep = odd.Solution.StepSize() / (3600)
        sub_losses = odd.Circuit.Losses()

        self.active_power.append((sub_losses[0]) * timestep / 1000)
        self.reactive_power.append((sub_losses[1]) * timestep / 1000)

    def get_metric(self):
        """Refer to base class for more details."""
        return {"active_power": self.active_power, "reactive_power": self.reactive_power}


class TimeseriesTotalPower(observer.MetricObserver):
    """Class for timeseries total power.

    Attributes:
        active_power (float): Time series active power
        reactive_power (float): Time series reactive power
    """

    def __init__(self):
        self.active_power = []
        self.reactive_power = []

    def compute(self):
        """Refer to base class for more details."""
        timestep = odd.Solution.StepSize() / (3600)
        sub_power = odd.Circuit.TotalPower()

        self.active_power.append((-sub_power[0]) * timestep / 1000)
        self.reactive_power.append((-sub_power[1]) * timestep / 1000)

    def get_metric(self):
        """Refer to base class for more details."""
        return {"active_power": self.active_power, "reactive_power": self.reactive_power}


class TimeseriesTotalPVPower(observer.MetricObserver):
    """Class for computing time series total pv power.

    Attributes:
        active_power (float): Time series active power
        reactive_power (float): Time series reactive power
    """

    def __init__(self):
        self.active_power = []
        self.reactive_power = []

    def compute(self):
        """Refer to base class for more details."""
        timestep = odd.Solution.StepSize() / (3600)
        pv_df = powerflow_results.get_pv_power_dataframe()
        if not pv_df.empty:
            pv_power = pv_df.sum().to_dict()
            self.active_power.append(pv_power["active_power"] * timestep / 1000)
            self.reactive_power.append(pv_power["reactive_power"] * timestep / 1000)
        else:
            self.active_power.append(0)
            self.reactive_power.append(0)

    def get_metric(self):
        """Refer to base class for more details."""
        return {"active_power": self.active_power, "reactive_power": self.reactive_power}


class TotalLossEnergy(observer.MetricObserver):
    """Class for computing total loss.

    Attributes:
        total_loss (float): Store for total energy
    """

    def __init__(self):
        self.total_loss = {"active_power": 0, "reactive_power": 0}

    def compute(self):
        """Refer to base class for more details."""
        timestep = odd.Solution.StepSize() / (3600)
        sub_losses = odd.Circuit.Losses()

        self.total_loss["active_power"] += (sub_losses[0]) * timestep / 1000000
        self.total_loss["reactive_power"] += (sub_losses[1]) * timestep / 1000000

    def get_metric(self):
        """Refer to base class for more details."""
        return self.total_loss


class TotalEnergy(observer.MetricObserver):
    """Class for computing total energy.

    Attributes:
        total_energy (float): Store for total energy
    """

    def __init__(self, export_only: bool = False, import_only: bool = False):
        self.total_energy = {"active_power": 0, "reactive_power": 0}
        self.export_only = export_only
        self.import_only = import_only

    def compute(self):
        """Refer to base class for more details."""
        timestep = odd.Solution.StepSize() / (3600)
        sub_power = odd.Circuit.TotalPower()

        if self.export_only and not self.import_only and sub_power[0] < 0:
            return

        if self.import_only and not self.export_only and sub_power[0] > 0:
            return

        self.total_energy["active_power"] += (-sub_power[0]) * timestep / 1000
        self.total_energy["reactive_power"] += (-sub_power[1]) * timestep / 1000

    def get_metric(self):
        """Refer to base class for more details."""
        return self.total_energy


class TotalPVGeneration(observer.MetricObserver):
    """Class for computing total energy in MWh.

    Attributes:
        pv_energy (float): Store for total energy
    """

    def __init__(self):
        self.pv_energy = {"active_power": 0, "reactive_power": 0}

    def compute(self):
        """Refer to base class for more details."""
        timestep = odd.Solution.StepSize() / (3600)
        pv_df = powerflow_results.get_pv_power_dataframe()
        if not pv_df.empty:
            pv_power = pv_df.sum().to_dict()
            self.pv_energy["active_power"] += pv_power["active_power"] * timestep / 1000
            self.pv_energy["reactive_power"] += pv_power["reactive_power"] * timestep / 1000

    def get_metric(self):
        """Refer to base class for more details."""
        return self.pv_energy


class SARDI_aggregated(observer.MetricObserver):
    """Class for computing SARDI aggregated metric.

    Attributes:
        loading_limit (ThermalLoadingLimit): Instance of `ThermalLoadingLimit`
            data model.
        sardi_transformer (float): SARDI_line metric
        counter (int): Counter for keeping track how number
            of times compute function is called
        network (nx.Graph): Networkx graph representing
            distribution network
    """

    def __init__(self, loading_limit: float = 1.0, voltage_limit: dict | None = None):
        """Constructor for `SARDI_line` class.

        Args:
            loading_limit (float): Per unit loading limit
                used for computing SARDI_line metric.
            voltage_limit (dict): Voltage threshold
        """
        voltage_limit = (
            {"overvoltage_threshold": 1.05, "undervoltage_threshold": 0.95}
            if voltage_limit is None
            else voltage_limit
        )

        self.loading_limit = data_model.ThermalLoadingLimit(threshold=loading_limit)
        self.voltage_limit = data_model.VoltageViolationLimit(**voltage_limit)
        self.sardi_aggregated = 0
        self.counter = 0

    def _get_initial_dataset(self):
        """Get initial dataset for computing the metric."""
        self.network = asset_metrics.networkx_from_opendss_model()
        self.load_bus_map = dss_util.get_bus_load_dataframe().set_index("busname")
        self.substation_bus = dss_util.get_source_node()
        self.bus_load_flag_df = dss_util.get_bus_load_flag()

    def compute(self):
        """Refer to base class for more details."""

        # Get line loading dataframe
        line_loading_df = powerflow_results.get_loading_dataframe()
        voltage_df = powerflow_results.get_voltage_dataframe()

        if not self.counter:
            self._get_initial_dataset()

        bus_with_voltage_violations = _get_voltage_impacted_buses(
            voltage_df,
            self.voltage_limit.overvoltage_threshold,
            self.voltage_limit.undervoltage_threshold,
        )

        overloaded_branches = line_loading_df.filter(
            pl.col("loading(pu)") > self.loading_limit.threshold
        )["branch"].to_list()
        total_impacted_load_buses = bus_with_voltage_violations
        if overloaded_branches:
            connected_buses = _get_unimpacted_buses(
                self.network, overloaded_branches, self.substation_bus
            )
            impacted_buses = self.bus_load_flag_df.loc[
                self.bus_load_flag_df.index.difference(connected_buses)
            ]
            impacted_load_buses = set(impacted_buses[impacted_buses["is_load"] == 1].index)
            total_impacted_load_buses = impacted_load_buses.union(bus_with_voltage_violations)

        total_load = odd.Loads.Count()
        affected_loads = list(
            set(self.load_bus_map.loc[list(total_impacted_load_buses)]["loadname"])
        )
        self.sardi_aggregated += len(affected_loads) * 100 / total_load
        self.counter += 1

    def get_metric(self):
        """Refer to base class for more details."""
        return {
            "sardi_aggregated": self.sardi_aggregated / self.counter if self.counter > 0 else 0
        }


class SARDI_line(observer.MetricObserver):
    """Class for computing SARDI line metric.

    Attributes:
        loading_limit (ThermalLoadingLimit): Instance of `ThermalLoadingLimit`
            data model.
        sardi_line (float): SARDI_line metric
        counter (int): Counter for keeping track how number
            of times compute function is called
        network (nx.Graph): Networkx graph representing
            distribution network
    """

    def __init__(self, loading_limit: float = 1.0):
        """Constructor for `SARDI_line` class.

        Args:
            loading_limit (float): Per unit loading limit
                used for computing SARDI_line metric.
        """

        self.loading_limit = data_model.ThermalLoadingLimit(threshold=loading_limit)
        self.sardi_line = 0
        self.counter = 0

    def _get_initial_dataset(self):
        """Get initial dataset for computing the metric."""
        self.network = asset_metrics.networkx_from_opendss_model()
        self.substation_bus = dss_util.get_source_node()
        self.bus_load_flag_df = dss_util.get_bus_load_flag()
        self.load_bus_map = dss_util.get_bus_load_dataframe().set_index("busname")

    def compute(self):
        """Refer to base class for more details."""

        # Get line loading dataframe
        line_loading_df = powerflow_results.get_loading_dataframe()

        if not self.counter:
            self._get_initial_dataset()

        overloaded_lines = line_loading_df.filter(
            pl.col("loading(pu)") > self.loading_limit.threshold
        )["branch"].to_list()

        if overloaded_lines:
            connected_buses = _get_unimpacted_buses(
                self.network, overloaded_lines, self.substation_bus
            )
            impacted_buses = self.bus_load_flag_df.loc[
                self.bus_load_flag_df.index.difference(connected_buses)
            ]
            impacted_load_buses = impacted_buses[impacted_buses["is_load"] == 1].index
            affected_loads = set(list(self.load_bus_map.loc[impacted_load_buses]["loadname"]))
            total_load = odd.Loads.Count()
            self.sardi_line += len(affected_loads) * 100 / total_load

        self.counter += 1

    def get_metric(self):
        """Refer to base class for more details."""
        return {"sardi_line": self.sardi_line / self.counter if self.counter > 0 else 0}


class SARDI_voltage(observer.MetricObserver):
    """Class for managing the computation of SARDI voltage metric.

    Attributes:
        upper_threshold (float): Voltage upper threshold
        lower_threshold (float): Voltage lower threshold
        sardi_voltage (float): SARDI voltage metric
        counter (int): Counter for keeping track how number
            of times compute function is called
        load_bus_map (pd.DataFrame): Dataframe containing mapping
            between bus and load
    """

    def __init__(self, upper_threshold: float = 1.05, lower_threshold: float = 0.95):
        """Constructor for SARDI_voltage metric.

        Args:
            upper_threshold (float): Voltage upper threshold
            lower_threshold (float): Voltage lower threshold
        """

        self.upper = upper_threshold
        self.lower = lower_threshold
        self.sardi_voltage = 0
        self.counter = 0

    def compute(self):
        """Refer to base class for more details."""

        # Get voltage dataframe and load bus mapper
        voltage_df = powerflow_results.get_voltage_dataframe()

        if not hasattr(self, "load_bus_map"):
            self.load_bus_map = dss_util.get_bus_load_dataframe().set_index("busname")

        # Filter voltages for load buses only
        # and count the number of overvoltages and undervoltages
        impacted_buses = _get_voltage_impacted_buses(voltage_df, self.upper, self.lower)
        affected_loads = list(set(self.load_bus_map.loc[impacted_buses]["loadname"]))

        self.sardi_voltage += len(affected_loads) * 100 / len(self.load_bus_map)
        self.counter += 1

    def get_metric(self):
        """Refer to base class for more details."""
        return {"sardi_voltage": self.sardi_voltage / self.counter if self.counter > 0 else 0}



""" This module implements commad line interface for exporting
schemas and handling opt in vscode update."""

from pathlib import Path
from typing import List

import click
from pydantic import BaseModel
import pydantic

from emerge.scenarios import data_model
from emerge.utils import util
from emerge.cli.timeseries_simulation import TimeseriesSimulationInput


class SchemaItemModel(BaseModel):
    name: str
    model: pydantic._internal._model_construction.ModelMetaclass

    class Config:
        arbitrary_types_allowed = True


class SchemaManager:
    def __init__(self, schema_folder: str = ".vscode"):
        self.schema_folder = Path(schema_folder)
        self.schemas: List[SchemaItemModel] = []

    def add_schema(self, name: str, model: BaseModel):
        self.schemas.append(SchemaItemModel(name=name, model=model))

    def generate_and_save_schemas(self):
        if not self.schema_folder.exists():
            self.schema_folder.mkdir()

        for schema in self.schemas:
            json_schema = schema.model.model_json_schema()
            schema_file = self.schema_folder / f"{schema.name}.json"
            util.write_file(json_schema, schema_file)

    def configure_vscode_settings(self):
        vscode_settings_file = self.schema_folder / "settings.json"

        if not vscode_settings_file.exists():
            util.write_file({}, vscode_settings_file)

        settings = util.read_file(vscode_settings_file, use_json5=True)
        vscode_key = "json.schemas"

        if vscode_key not in settings:
            settings[vscode_key] = []

        for schema in self.schemas:
            updated_in_place = False
            schema_file = self.schema_folder / f"{schema.name}.json"

            for item in settings[vscode_key]:
                if schema.name in item.get("url", ""):
                    item["url"] = str(schema_file)
                    updated_in_place = True

            if not updated_in_place:
                settings[vscode_key].append({"fileMatch": ["*.json"], "url": str(schema_file)})

        util.write_file(settings, vscode_settings_file)


@click.command()
@click.option(
    "-vc",
    "--vscode",
    default=False,
    show_default=True,
    help="""Update JSON schemas in vscode settings. Note will create .vscode folder
    if not present in the current directory.""",
)
def create_schemas(vscode: bool):
    """Function to handle the JSON schemas for emerge package."""

    schema_manager = SchemaManager()
    schema_manager.add_schema("emerge_scenario_schema", data_model.DERScenarioConfigModel)
    schema_manager.add_schema("emerge_timeseries_simulation_schema", TimeseriesSimulationInput)
    schema_manager.generate_and_save_schemas()
    if vscode:
        schema_manager.configure_vscode_settings()



""" Module for computing timeseries line loading stats. """

from typing import Dict, List
import numpy as np

import polars
import opendssdirect as odd

from emerge.metrics import observer
from emerge.simulator import powerflow_results


def get_line_loading_df():
    return powerflow_results.get_loading_dataframe()


class OverloadedLines(observer.MetricObserver):
    """Class for computing time series loading metrics for lines."""

    def __init__(self):
        self.metrics = {}

    def compute(self) -> None:
        df = powerflow_results.get_loading_dataframe()
        loading_dict = dict(zip(df["branch"], df["loading(pu)"]))
        if len(self.metrics) == 0:
            self.metrics = {key: [] for key in loading_dict}

        for line, metric in loading_dict.items():
            self.metrics[line].append(metric)

    def get_metric(self) -> Dict:
        overloaded_lines = {}
        for line, loading_arr in self.metrics.items():
            loading_arr_np = np.array(loading_arr)
            if len(loading_arr_np[loading_arr_np > 1.0]):
                overloaded_lines[line] = loading_arr

        return overloaded_lines


class LineLoadingBins(observer.MetricObserver):
    """Class for computing line loading distribution bins."""

    def __init__(self, bins: List[float]):
        bins = list(set(bins))
        bins.sort()

        if len(bins) < 2:
            raise ValueError(f"{bins} should at least have two unique items.")

        self.metrics = {f">{bins[id]}__<{el}": 0 for id, el in enumerate(bins[1:])}
        self.metrics[f"<{bins[0]}"] = 0
        self.metrics[f">{bins[-1]}"] = 0

    def compute(self) -> None:
        df = get_line_loading_df()
        timestep = odd.Solution.StepSize() / (3600)

        for metric in self.metrics:
            if "__" in metric:
                low_val = float(metric.split("__")[0].split(">")[1])
                high_val = float(metric.split("__")[1].split("<")[1])

                self.metrics[metric] += (
                    len(
                        df.filter(
                            (polars.col("loading(pu)") > low_val)
                            & (polars.col("loading(pu)") < high_val)
                        )
                    )
                    * timestep
                )

            elif "<" in metric and "__" not in metric:
                low_val = float(metric.split("<")[1])
                self.metrics[metric] += (
                    len(df.filter((polars.col("loading(pu)") < low_val))) * timestep
                )

            elif ">" in metric and "__" not in metric:
                high_val = float(metric.split(">")[1])
                self.metrics[metric] += (
                    len(df.filter((polars.col("loading(pu)") > high_val))) * timestep
                )

    def get_metric(self) -> Dict:
        return self.metrics


class LineLoadingStats(observer.MetricObserver):
    """Class for computing line loading statistics."""

    def __init__(self):
        self.metrics = {
            metric: []
            for metric in [
                "min",
                "max",
                "median",
                "mean",
                "quantile_0.1",
                "quantile_0.25",
                "quantile_0.75",
                "quantile_0.90",
            ]
        }

    def compute(self) -> None:
        df = get_line_loading_df()  # noqa
        for metric in self.metrics:
            if "_" not in metric:
                self.metrics[metric].append(eval(f"df.{metric}()['loading(pu)'][0]"))
            else:
                param = float(metric.split("_")[1])
                self.metrics[metric].append(
                    eval(f"df.{metric.split('_')[0]}({param})['loading(pu)'][0]")
                )

    def get_metric(self) -> Dict:
        return self.metrics



""" Module for computing time series metrics for multi scenario simulation."""
from pathlib import Path
import multiprocessing
from typing import Annotated
import json

import click
from emerge.cli import get_num_core, get_observers
from emerge.cli.timeseries_simulation import TimeseriesSimulationInput

from emerge.metrics import observer
from emerge.simulator import simulation_manager
from emerge.simulator.opendss import OpenDSSSimulator
from pydantic import Field


class ScenarioTimeseriesSimulationInput(TimeseriesSimulationInput):
    scenario_file: Annotated[Path, Field(..., description="Path to .dss file to load")]


def _run_timeseries_sim(config: ScenarioTimeseriesSimulationInput):
    opendss_instance = OpenDSSSimulator(config.master_dss_file)
    opendss_instance.post_redirect(config.scenario_file.absolute())

    manager = simulation_manager.OpenDSSSimulationManager(
        opendss_instance=opendss_instance,
        simulation_start_time=config.start_time,
        profile_start_time=config.profile_start_time,
        simulation_end_time=config.end_time,
        simulation_timestep_min=config.resolution_min,
    )

    subject = observer.MetricsSubject()
    observers = get_observers.get_observers(config.metrics.model_dump())

    for _, observer_ in observers.items():
        subject.attach(observer_)

    manager.simulate(subject)

    export_folder = Path(config.export_path) / config.scenario_file.stem
    export_folder.mkdir(exist_ok=True, parents=True)
    manager.export_convergence(export_folder / "convergence_report.csv")
    observer.export_csv(list(observers.values()), export_folder)


@click.command()
@click.option(
    "-c",
    "--config",
    help="Path to config file for generating scenarios",
)
@click.option(
    "-sf",
    "--scenario_folder",
    help="Path to scenario folder",
)
@click.option(
    "-nc",
    "--num-core",
    default=1,
    show_default=True,
    help="Num of cores to run simulation with.",
)
def multi_timeseries_simulation(config: str, scenario_folder: str, num_core: int):
    """Run multiscenario time series simulation and compute
    time series metrics."""
    with open(config, "r", encoding="utf-8") as file:
        config_dict = json.load(file)

    config = TimeseriesSimulationInput.model_validate(config_dict)

    scenario_folder = Path(scenario_folder)
    timeseries_input = []
    for file_path in scenario_folder.iterdir():
        if file_path.suffix == ".dss":
            timeseries_input.append(
                ScenarioTimeseriesSimulationInput(**config.model_dump(), scenario_file=file_path)
            )

    num_core = get_num_core(num_core, len(timeseries_input))

    if num_core > 0:
        with multiprocessing.Pool(int(num_core)) as p:
            p.map(_run_timeseries_sim, timeseries_input)



""" Module for computing timeseries voltage stats. """

from typing import List

import opendssdirect as odd
import polars

from emerge.metrics import observer
from emerge.simulator import powerflow_results


def get_voltage_df():
    return powerflow_results.get_voltage_dataframe()


class NodeVoltageTimeSeries(observer.MetricObserver):
    def __init__(self):
        self.metrics = {}

    def compute(self) -> None:
        df = powerflow_results.get_voltage_dataframe()
        if not self.metrics:
            self.metrics = {busname: [] for busname in df.index}
        for key, value in df.to_dict()["voltage(pu)"].items():
            self.metrics[key].append(value)

    def get_metric(self) -> dict:
        return self.metrics


class NodeVoltageStats(observer.MetricObserver):
    """Class for computing node voltage statistical metrics."""

    def __init__(self):
        self.metrics = {
            metric: []
            for metric in [
                "min",
                "max",
                "median",
                "mean",
                "quantile_0.1",
                "quantile_0.25",
                "quantile_0.75",
                "quantile_0.90",
            ]
        }

    def compute(self) -> None:
        df = get_voltage_df()  # noqa
        for metric in self.metrics:
            if "_" not in metric:
                self.metrics[metric].append(eval(f"df.{metric}()['voltage(pu)'][0]"))
            else:
                param = float(metric.split("_")[1])
                self.metrics[metric].append(
                    eval(f"df.{metric.split('_')[0]}({param})['voltage(pu)'][0]")
                )

    def get_metric(self) -> dict:
        return self.metrics


class NodeVoltageBins(observer.MetricObserver):
    """Class for computing node voltage statistical metrics."""

    def __init__(self, bins: List[float]):
        bins = list(set(bins))
        bins.sort()

        if len(bins) < 2:
            raise ValueError(f"{bins} should at least have two unique items.")

        self.metrics = {f">{bins[id]}__<{el}": 0 for id, el in enumerate(bins[1:])}
        self.metrics[f"<{bins[0]}"] = 0
        self.metrics[f">{bins[-1]}"] = 0

    def compute(self) -> None:
        df = get_voltage_df()
        timestep = odd.Solution.StepSize() / (3600)

        for metric in self.metrics:
            if "__" in metric:
                low_val = float(metric.split("__")[0].split(">")[1])
                high_val = float(metric.split("__")[1].split("<")[1])

                self.metrics[metric] += (
                    len(
                        df.filter(
                            (polars.col("voltage(pu)") > low_val)
                            & (polars.col("voltage(pu)") < high_val)
                        )
                    )
                    * timestep
                )

            elif "<" in metric and "__" not in metric:
                low_val = float(metric.split("<")[1])
                self.metrics[metric] += (
                    len(df.filter((polars.col("voltage(pu)") < low_val))) * timestep
                )

            elif ">" in metric and "__" not in metric:
                high_val = float(metric.split(">")[1])
                self.metrics[metric] += (
                    len(df.filter((polars.col("voltage(pu)") > high_val))) * timestep
                )

    def get_metric(self) -> dict:
        return self.metrics



"""Command line utility for EMeRGE."""


import click

from emerge.network import feeder_geojson
from emerge.simulator import opendss
from emerge.cli.scenario import generate_scenarios
from emerge.cli.multiscenario_metrics import (
    multi_timeseries_simulation,
)
from emerge.cli.schema_generator import create_schemas
from emerge.cli.timeseries_simulation import timeseries_simulation
from emerge.cli.nodal_hosting_capacity import nodal_hosting_analysis


@click.command()
@click.option(
    "-m",
    "--master-file",
    help="Path to master dss file",
)
@click.option(
    "-o",
    "--output-folder",
    default="./geojsons",
    show_default=True,
    help="Ouput directory for storing the geojsons",
)
def create_geojsons(master_file, output_folder):
    """Command line function to generate geojsons from opendss model"""

    opendss_instance = opendss.OpenDSSSimulator(master_file)
    feeder_geojson.create_feeder_geojson(opendss_instance.dss_instance, output_folder)


@click.group()
def cli():
    """Entry point"""


cli.add_command(timeseries_simulation)
cli.add_command(create_geojsons)
cli.add_command(generate_scenarios)
cli.add_command(multi_timeseries_simulation)
cli.add_command(create_schemas)
cli.add_command(nodal_hosting_analysis)



import os


def get_num_core(num_core: int, num_inputs: int) -> int:
    """Function to get actual number of core."""

    if num_core > os.cpu_count():
        num_core = os.cpu_count() - 1 or os.cpu_count()

    if num_inputs < num_core:
        num_core = num_inputs

    return num_core
